//////////////////////////////////////////////////////////////////////////////////////////////
// (C)2005 Vaclav Gassenbauer, Graphical systems course, CTU FEL, Prague
// Modifications:
// 2007  Petr Felkel, corrections of weights handling 
//       - glRasterPos ignores w / divided manually
//		 - demos did not premultiplied by w
//		 - 4f versions of positions used instead of 2f
//		 - different weight input
// TODO - preset knot vector values (double-, tripple-points, group move,... )
//////////////////////////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>
#include <GL/glut.h>

#include "graph_util.h"

static int mainWindow, worldWindow, knotWindow;		//instance podoken
#define WINDOW_TITLE    "OpenGL NURBS demonstration"// titulek okna
#define WINDOW_WIDTH    900	                        // velikost okna
#define WINDOW_HEIGHT   600
#define GAP  8										// mezery mezi okny

static int mainWidth, mainHeight;
static GLfloat scaleScene = 1, angleScene = 0;
static GLfloat translScene[] = { .0f, .0f };


GLint modifiers, mouseStatus;
static bool mainActive = true;


typedef struct {
	GLfloat x, y, z, w; } vertex4f;

#define CTRL_POINT_MAX 42
#define KNOT_MAX 50

GLUnurbs *nurbs;
vertex4f ctrlPoints[CTRL_POINT_MAX *sizeof(vertex4f)] =
	{	{-2, -1, 0, 1}, {-1, 1, 0, 1}, { 1, 1.5, 0, 1},
		{ 0, -1, 0, 1}, { 2, 0, 0, 1}	};
GLfloat knotVector[KNOT_MAX] =
	{ .0, .0, .0,   .33, .67,   1., 1., 1. };

int numberOfCtrlPoints = 5;
int numberOfKnots = 8;
int order = 3;
GLfloat sampleTolerance = 0.01;


#define FONT1 GLUT_BITMAP_8_BY_13
#define FONT1_W 8
#define FONT1_H 13
GLfloat BLACK[]		= {0, .0, 0, 1},
		COMMENT[]	= {0, .6, 0, 1}, 
		ERR[]		= {1,  0, 0, 1},
		RED[]		= {1,  0, 0, 1},
		HIT[]		= {1,  1, 0, 1};


void reDrawAll(void)
{
	glutSetWindow(knotWindow);
		glutPostRedisplay();
	glutSetWindow(worldWindow);
		glutPostRedisplay();
	glutSetWindow(mainWindow);
}


#include "worldWindow.h"
#include "knotWindow.h"

//---------------------------------------------------------------------
// Callback funkce volana pri vyskytu chyby pri renderovani NURB krivek
//---------------------------------------------------------------------
char errMessage[][100] = { {""}, {""}, {""} };
void __stdcall onError(int errorCode)
{
	static line = 0;

	memmove(errMessage, &errMessage[1], 2*sizeof(errMessage[0]));

	sprintf(errMessage[2], "line:%3d     code:%8d   -%s\n", line++, errorCode, gluErrorString(errorCode));

	printf("%s\n", errMessage[2]);

	int activeWindow = glutGetWindow();
	glutSetWindow(mainWindow);
		glutPostRedisplay();
	glutSetWindow(activeWindow);
}


 


void mainInit(void) 
{
    nurbs = gluNewNurbsRenderer();                    // vytvoreni NURBS
	gluNurbsCallback(nurbs, GLU_ERROR, (void(__stdcall*)(void))onError);
	gluNurbsProperty(nurbs, GLU_SAMPLING_TOLERANCE, 1.0); 

}


//---------------------------------------------------------------------
// Funkce pro rozmisteni podoken na plose
//---------------------------------------------------------------------
static void mainReshape(int width,  int height) 
{
	mainWidth = width; mainHeight = height;
    glViewport(0, 0, width, height); 
    glClearColor(0.8, 0.8, 0.8, 1.0); 
	glClearDepth(1.);

    glViewport(0, 0, width, height);
    glMatrixMode(GL_PROJECTION);
	    glLoadIdentity();
		gluOrtho2D(0, width, height, 0);
    glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();

	glutSetWindow(worldWindow);
    glutPositionWindow(GAP, GAP);
    glutReshapeWindow((int)((width -GAP*3) *(5.5/10)), 
		(int)((height-GAP*3) *(9.5/10)));

	glutSetWindow(knotWindow);
    glutPositionWindow(	width  -(  GAP +(width  -3*GAP)*4.5/10),
						height -(2*GAP +(height -4*GAP)*4.5/10));
	glutReshapeWindow(	(int)((width  -3*GAP)*(4.5/10)), 
						(int)((height -4*GAP)*(4. /10)));

	glutSetWindow(mainWindow);

}
 
static void mainDisplay(void)
{
	static run = 0;
	printf("run %d\n", run++);
	int x = mainWidth  -GAP -(mainWidth  -GAP*3)*(4.5/10) 
			+((mainWidth -GAP*3)*(4.5/10) -(48*FONT1_W)) / 2;
	int y = GAP +FONT1_H +((mainHeight -GAP*4)*(5./10) -(23*FONT1_H)) / 2;

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); // vymazani barvoveho a hloubkoveho bufferu

	printfGlutBitmapFont(x, y += FONT1_H,BLACK,   FONT1, "GLUnurbs *nurbs");
	printfGlutBitmapFont(x, y += FONT1_H,BLACK,   FONT1, "GLfloat knotVector[] = { a1, a2, ... , a3 };   ");
	printfGlutBitmapFont(x, y += FONT1_H,BLACK,   FONT1, "GLfloat controlPoints[] = {{x1, y2, z1, w1},   ");
	printfGlutBitmapFont(x, y += FONT1_H,BLACK,   FONT1, "        ..., {xn, yn, zn, wn} };               ");
	printfGlutBitmapFont(x, y += FONT1_H,COMMENT, FONT1, "//numberOfKnot == order +                      ");
	printfGlutBitmapFont(x, y += FONT1_H,BLACK,   FONT1, "//      +numberOfCtrlPoints                    ");
	printfGlutBitmapFont(x, y+=2*FONT1_H,BLACK,   FONT1, "   ...                                         ");
	printfGlutBitmapFont(x, y += FONT1_H,BLACK,   FONT1, "nurbs = gluNewNurbsRenderer();                 ");
	printfGlutBitmapFont(x, y += FONT1_H,BLACK,   FONT1, "gluNurbsCallback(nurbs, GLU_ERROR,             ");
	printfGlutBitmapFont(x, y += FONT1_H,BLACK,   FONT1, "        void (*fn)(void));");
	printfGlutBitmapFont(x, y += FONT1_H,BLACK,   FONT1, "gluNurbsProperty(nurbs, param, value);         ");

	printfGlutBitmapFont(x, y+=2*FONT1_H,BLACK,   FONT1, "gluBeginCurve(nurbs);                          ");
	printfGlutBitmapFont(x, y += FONT1_H,BLACK,   FONT1, "    gluNurbsCurve(nurbs,                       ");
	printfGlutBitmapFont(x+28*FONT1_W, y,COMMENT, FONT1,                           "//object of NURBS    ");
	printfGlutBitmapFont(x, y += FONT1_H,BLACK,   FONT1, "        numberOfKnots,                         ");
	printfGlutBitmapFont(x, y += FONT1_H,BLACK,   FONT1, "        knotVector,                            ");
	printfGlutBitmapFont(x+28*FONT1_W, y,COMMENT, FONT1,                           "//pointer to array   ");
	printfGlutBitmapFont(x, y += FONT1_H,BLACK,   FONT1, "        stride,                                ");
	printfGlutBitmapFont(x+16*FONT1_W, y,COMMENT, FONT1,              "//4 GLfloat between controlPoints ");
	printfGlutBitmapFont(x, y += FONT1_H,BLACK,   FONT1, "        controlPoints,                         ");
	printfGlutBitmapFont(x+28*FONT1_W, y,COMMENT, FONT1,                           "//pointer to array   ");
	printfGlutBitmapFont(x, y += FONT1_H,BLACK,   FONT1, "        order,                                 ");
	printfGlutBitmapFont(x+28*FONT1_W, y,COMMENT, FONT1,                           "//degree +1          ");
	printfGlutBitmapFont(x, y += FONT1_H,BLACK,   FONT1, "        GL_MAP1_VERTEX_4);                     ");
	printfGlutBitmapFont(x+28*FONT1_W, y,COMMENT, FONT1,                           "//type of NURBS      ");
	printfGlutBitmapFont(x, y += FONT1_H,BLACK,   FONT1, "gluEndCurve(nurbs);                            ");
	printfGlutBitmapFont(x, y += FONT1_H,BLACK,   FONT1, "gluDeleteNurbsRenderer(nurbs);                 ");

	printfGlutBitmapFont(2*GAP, mainHeight-2*FONT1_H, ERR,     FONT1, "error message:   ");
	printfGlutBitmapFont(2*GAP, mainHeight-1*FONT1_H, COMMENT, FONT1, " (gluErrorString)");
	printfGlutBitmapFont(180, mainHeight -5-2*FONT1_H, ERR, FONT1, "%s", errMessage[0]);
	printfGlutBitmapFont(180, mainHeight -5-1*FONT1_H, ERR, FONT1, "%s", errMessage[1]);
	printfGlutBitmapFont(180, mainHeight -5-0*FONT1_H, ERR, FONT1, "%s", errMessage[2]);

    glFlush();                                      // provedeni a vykresleni vsech zmen
    glutSwapBuffers();
}

void mainEntryFunc(int state) 
{
	if (state == GLUT_ENTERED) mainActive = true;
	else mainActive = false;
}


//---------------------------------------------------------------------
// Tato funkce je volana pri stlaceni ASCII klavesy
//---------------------------------------------------------------------
void mainKeyboard(unsigned char key, int x, int y)
{
    if (key>='A' && key<='Z')                       // uprava velkych pismen na mala
        key+=(unsigned char)('a'-'A');              // pro zjednoduseni prikazu switch

    switch (key) {
        case 27:    exit(0);            break;      // ukonceni aplikace
        case 'q':   exit(0);            break;      // ukonceni aplikace
        case 'f':   glutFullScreen();   break;      // prepnuti na celou obrazovku
        case 'w':   
			glutSetWindow(mainWindow);
			glutReshapeWindow(WINDOW_WIDTH, WINDOW_HEIGHT); 
		    glutInitWindowPosition(30, 30);                 // pocatecni pozice leveho horniho rohu okna
			break;
        default:                        break;
    }
}

 
//---------------------------------------------------------------------
// Hlavni funkce konzolove aplikace
//---------------------------------------------------------------------
int main(int argc, char** argv)
{
    glutInit(&argc, argv);                          // inicializace knihovny GLUT
    glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH | GLUT_STENCIL);    // nastaveni dvou barvovych bufferu a pameti hloubky
    glutInitWindowPosition(30, 30);                 // pocatecni pozice leveho horniho rohu okna
    glutInitWindowSize(WINDOW_WIDTH, WINDOW_HEIGHT);// pocatecni velikost okna

    mainWindow = glutCreateWindow(WINDOW_TITLE);                 // vytvoreni okna pro kresleni
		glutReshapeFunc(mainReshape);                      // registrace funkce volane pri zmene velikosti okna
		glutDisplayFunc(mainDisplay);                     // registrace funkce volane pri prekreslovani okna
		glutEntryFunc(mainEntryFunc);
	    glutKeyboardFunc(mainKeyboard);                   // registrace funkce volane pri stlaceni klavesy
//    glutMouseFunc(mainMouse);                         // registrace funkce volane pri stisku tlacitka mysi
//    glutMotionFunc(mainMouseMotion);                  // registrace funkce volane pri pohybu mysi

	mainInit();
	worldInit();
	knotInit();
	glutMainLoop();                                 // nekonecna smycka, kde se volaji zaregistrovane funkce
    return 0;                                       // navratova hodnota vracena operacnimu systemu
}
 