#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "graph_util.h"


//---------------------------------------------------------------------
// Tato funkce vykresli retezec zadanym bitmapovym fontem
//---------------------------------------------------------------------
void printfGlutBitmapFont(void *font, char *format, ...)
{
	va_list args;
	char buffer[255], *s = buffer;

	va_start(args, format);
	vsprintf(buffer, format, args);
	va_end(args);

    while (*s) glutBitmapCharacter(font, *s++);    
}

void printfGlutBitmapFont(int x, int y, GLfloat *color, void *font, char *format, ...)
{
	va_list args;
	char buffer[255], *s = buffer;

	va_start(args, format);
	vsprintf(buffer, format, args);
	va_end(args);

	glColor4fv(color);
	glRasterPos2f(x, y);
    while (*s) glutBitmapCharacter(font, *s++);    
}


//---------------------------------------------------------------------
// Tato funkce vykresli retezec zadanym stroke fontem
//---------------------------------------------------------------------
void printfGlutStrokeFont(GLfloat *rot, GLfloat *scale, void *font, char *format, ...)
{
	va_list args;
	char buffer[255], *s = buffer;

	va_start(args, format);
	vsprintf(buffer, format, args);
	va_end(args);

	glPushMatrix();
	//glRotatef(rot[0], rot[1], rot[2], rot[3]);
	
	glRotatef(30, 0, 0, 1);
	while (*s) {
		glPushMatrix();
			glScalef(scale[0], scale[1], scale[2]);
			glutStrokeCharacter(font, *s++);    
		glPopMatrix();
		glTranslatef(0.5, 0, 0);
	}

	glPopMatrix();
}


/* glmCross: compute the cross product of two vectors
 *
 * u - array of 3 GLfloats (GLfloat u[3])
 * v - array of 3 GLfloats (GLfloat v[3])
 * n - array of 3 GLfloats (GLfloat n[3]) to return the cross product in
 */
GLvoid crossProduct(GLfloat* u, GLfloat* v, GLfloat* n)
{
  n[0] = u[1]*v[2] - u[2]*v[1];
  n[1] = u[2]*v[0] - u[0]*v[2];
  n[2] = u[0]*v[1] - u[1]*v[0];
}
 
/* glmNormalize: normalize a vector
 *
 * v - array of 3 GLfloats (GLfloat v[3]) to be normalized
 */
void normalize(GLfloat* v)
{
  GLfloat l;

  l = (GLfloat)sqrt(v[0]*v[0] + v[1]*v[1] + v[2]*v[2]);
  v[0] /= l;
  v[1] /= l;
  v[2] /= l;
}
 
//---------------------------------------------------------------------
// Funkce pro vynasobeni vektoru skalarem
//---------------------------------------------------------------------
void multVectByScalar(GLfloat* v, GLfloat scalar)
{
  v[0] *= scalar;
  v[1] *= scalar;
  v[2] *= scalar;
}


//---------------------------------------------------------------------
// Pretransformovani vrcholu matici matrix[16]
//---------------------------------------------------------------------
void transfVertex(GLdouble *matrix, GLfloat *vertex) {
	GLfloat res[] = { .0f, .0f, .0f , 0.f};

	register short i, k;

	for (i = 0; i < DIM; i++) 
		for (k = 0; k < DIM; k++)
			res[i] += (float)matrix[k*DIM +i] * vertex[k];

	memcpy(vertex, res, DIM * sizeof(GLfloat)); 
} 

//---------------------------------------------------------------------
// Upravena pomocna funkce pro aplikaci zpetne projekce. Pro bod zadany
// v souradnicich obrazovky [x, y] se vypocte odpovidajici bod ve
// svetovych souradnicich [xw, yw, zw]. Z-ova souradnice bodu na obra-
// zovce je vypoctena z hloubky ulozene ve framebufferu.
//---------------------------------------------------------------------
void unProject3D(GLfloat x, GLfloat y,
				 GLfloat *xw, GLfloat *yw, GLfloat *zw)
{    
	GLdouble xxw, yyw, zzw;

    GLint viewPort[4];				// pole pro ulozeni aktualne nastaveneho ViewPortu
    GLdouble modelViewMatrix[16];	// pole pro ulozeni aktualne nastavene matice ModelView
    GLdouble projectionMatrix[16];	// pole pro ulozeni aktualne nastavene matice Projection
    
    glGetIntegerv(GL_VIEWPORT, viewPort);
    glGetDoublev(GL_MODELVIEW_MATRIX, modelViewMatrix);
    glGetDoublev(GL_PROJECTION_MATRIX, projectionMatrix);
    
     // poecist hloubku pixelu primo z framebufferu
    // z pameti hloubky (Z-bufferu, depth bufferu)
    GLfloat z;			// promenna, do ktere se z framebufferu	nacte z-ova souradnice pixelu
    glReadPixels((GLint)x, (GLint)y, // pozice pixmapy
                 1, 1,               // rozmery pixmapy
                 GL_DEPTH_COMPONENT, // atribut fragmentu
                 GL_FLOAT,           // datovy format
                 &z                  // vysledna "pixmapa"
    );

	// provest vlastni zpetnou projekci
    gluUnProject(x, y, (GLdouble)z,
                 modelViewMatrix, projectionMatrix,
                 viewPort,
                 &xxw, &yyw, &zzw);

	*xw = (float) xxw; *yw = (float) yyw; *zw = (float) zzw;
}

//---------------------------------------------------------------------
// Upravena pomocna funkce pro aplikaci zpetne projekce. Pro bod zadany
// v souradnicich obrazovky [x, y] se vypocte odpovidajici bod ve
// svetovych souradnicich [xw, yw, zw]. Z-ova souradnice bodu na obra-
// zovce je vypoctena z hloubky ulozene ve framebufferu.
//---------------------------------------------------------------------
void unProject2D(GLfloat x, GLfloat y,
				 GLfloat *xw, GLfloat *yw)
{    
	GLdouble xxw, yyw, zzw;

    GLint viewPort[4];				// pole pro ulozeni aktualne nastaveneho ViewPortu
    GLdouble modelViewMatrix[16];	// pole pro ulozeni aktualne nastavene matice ModelView
    GLdouble projectionMatrix[16];	// pole pro ulozeni aktualne nastavene matice Projection
    
    glGetIntegerv(GL_VIEWPORT, viewPort);
    glGetDoublev(GL_MODELVIEW_MATRIX, modelViewMatrix);
    glGetDoublev(GL_PROJECTION_MATRIX, projectionMatrix);
    
	// provest vlastni zpetnou projekci
    gluUnProject(x, y, .0,
                 modelViewMatrix, projectionMatrix,
                 viewPort,
                 &xxw, &yyw, &zzw);

	*xw = (float) xxw; *yw = (float) yyw; 
}

