//////////////////////////////////////////////////////////////////////////////////////////////
// (C)2005 Vaclav Gassenbauer, Graphical systems course, CTU FEL, Prague
// Modifications:
// 2007  Petr Felkel, corrections of weights handling 
//       - glRasterPos ignores w / divided manually
//		 - demos did not premultiplied by w
//		 - 4f versions of positions used instead of 2f
//		 - different weight input
//////////////////////////////////////////////////////////////////////////////////////////////


typedef enum { ADD_POINT, DELETE_POINT, EDIT_MODE } tOperations;

tOperations operations = EDIT_MODE;


#define WORLD_MESH	10
static GLint selectId = 0;
static bool drawCtrlPolygon = true;
static int worldHitId = -1;
static bool worldActive = true;

static int worldWidth = 480, worldHeight = 480;
static int worldMouseX = 0, worldMouseY = 0;


void addCtrlPoint(GLfloat x, GLfloat y, GLfloat z, GLfloat w) 
{
	ctrlPoints[numberOfCtrlPoints].x = x;
	ctrlPoints[numberOfCtrlPoints].y = y;
	ctrlPoints[numberOfCtrlPoints].z = z;
	ctrlPoints[numberOfCtrlPoints].w = w;

	int min, max;
	for (min = 0;   knotVector[0] >= knotVector[min]; min++);
	for (max = min; knotVector[max] <  knotVector[numberOfCtrlPoints +order -1]; max++);
	GLfloat pomer = (GLfloat)(max -min +1) / (max -min +2);
	for ( ; min <= max; min++)
		knotVector[min] *= pomer;
	
	numberOfKnots = ++numberOfCtrlPoints +order;
	knotVector[numberOfKnots -1] = 1.f;
}

void moveCtrlPoint(int hit, GLfloat x, GLfloat y, GLfloat z, GLfloat w) 
{
	ctrlPoints[hit].x = x * w;
	ctrlPoints[hit].y = y * w;
	ctrlPoints[hit].z = z * w;
	ctrlPoints[hit].w = w;
}

void deleteCtrlPoint(int hit) 
{
	memmove(&ctrlPoints[hit], &ctrlPoints[hit+1], 
		(numberOfCtrlPoints-- -hit) *sizeof(vertex4f));
	memmove(&knotVector[hit], &knotVector[hit+1],
		(numberOfKnots-- -hit) *sizeof(GLfloat));
	worldHitId = -1;
}


void drawNurbs(GLenum mode) 
{
	GLubyte bitmap[] = { 0xF0, 0x90, 0x90, 0xF0 };    // ctverecek 8x4 pixely
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
	for (int i = 0; i < numberOfCtrlPoints; i++) {    // ridici body (ctverecek +text )
		glColor4f(1, (GLfloat)i /(numberOfKnots-order+1), 0, 1);
		if (mode == GL_SELECT) 
			glLoadName (i); 
		//glRasterPos4f(ctrlPoints[i].x,ctrlPoints[i].y,ctrlPoints[i].z,ctrlPoints[i].w); // tohle nechodi, musim "RUCNE"
		glRasterPos4f(ctrlPoints[i].x/ctrlPoints[i].w,ctrlPoints[i].y/ctrlPoints[i].w,ctrlPoints[i].z/ctrlPoints[i].w,1.0);
		glBitmap(4, 4, 2, 2, -5, -14, bitmap);
		if (mode == GL_RENDER) 
			printfGlutBitmapFont(GLUT_BITMAP_HELVETICA_10, "V%d", i);
	}

	if (drawCtrlPolygon && (mode == GL_RENDER)) {    // ridici polygon
		glColor4f(.4, .4, 0, 1);
		glBegin(GL_LINE_STRIP);
			for (int i = 0; i < numberOfCtrlPoints; i++) 
				//glVertex2f(ctrlPoints[i].x, ctrlPoints[i].y);
				glVertex4f(ctrlPoints[i].x, ctrlPoints[i].y, ctrlPoints[i].z, ctrlPoints[i].w);
		glEnd();
	}

	if (mode == GL_RENDER) {
		glColor3f(0.0f, 1.0f, 0.0f);
		gluBeginCurve(nurbs);							// zacatek specifikace NURBS
		gluNurbsCurve(nurbs,
	        numberOfCtrlPoints +order, knotVector,		// pocet slozek v uzlovem vektoru
			sizeof(vertex4f)/sizeof(GLfloat),			// pocet floatu pro jeden ridici bod
			(GLfloat*)ctrlPoints,									// ridici body
			order,										// stupen krivky
			GL_MAP1_VERTEX_4							// funkce ridicich bodu
		);
		gluEndCurve(nurbs);								// konec specifikace NURBS
	}

	if ((worldHitId != -1) && (mode == GL_RENDER)) {   // aktualni vaha w=
		static GLubyte buffer[] = { 0 };
		glColor4f(1, (GLfloat)worldHitId /(numberOfKnots-order+1), 0, 1);
		//glRasterPos2f(ctrlPoints[worldHitId].x, ctrlPoints[worldHitId].y);
		glRasterPos4f(ctrlPoints[worldHitId].x/ctrlPoints[worldHitId].w,ctrlPoints[worldHitId].y/ctrlPoints[worldHitId].w,ctrlPoints[worldHitId].z/ctrlPoints[worldHitId].w,1.0);
		glBitmap(0, 0, 0, 0, -10, 6, buffer);
	
		printfGlutBitmapFont(GLUT_BITMAP_HELVETICA_10, 
			"w=%6.2f", ctrlPoints[worldHitId].w);
	}

}


/*  pickSquares() sets up selection mode, name stack, 
 *  and projection matrix for picking.  Then the 
 *  objects are drawn.
 */
#define BUFSIZE 512

int pickNurbs(void)
{
	GLuint selectBuf[BUFSIZE];
	GLint hits;
	GLint viewport[4];

	glGetIntegerv (GL_VIEWPORT, viewport);

	glSelectBuffer (BUFSIZE, selectBuf);
	glRenderMode (GL_SELECT);

	glInitNames();
	glPushName(-1);

	glMatrixMode (GL_PROJECTION);
	glPushMatrix ();
	glLoadIdentity ();
	/*  create 5x5 pixel picking region near cursor location	*/
	gluPickMatrix ((GLdouble) worldMouseX-0, (GLdouble) (viewport[3] - (worldMouseY-0)), 
					6.0, 6.0, viewport);
	gluOrtho2D(-4, 4, -4, 4);
	drawNurbs (GL_SELECT);

	glMatrixMode (GL_PROJECTION);
	glPopMatrix ();
	glFlush ();

	hits = glRenderMode (GL_RENDER);
	printf("%d\n", hits);

	worldHitId = (0 < hits)? selectBuf[3] :-1;
	return hits;
//	processHits (hits, selectBuf);
//	glutPostRedisplay();
} 
 
static void worldDisplay(GLenum mode)
{
	if ((mode == GL_RENDER) && (selectId == 0)) {
		glColor4f(.3f, .3f, .3f, 1.f);
		//glEnable(GL_LINE_STIPPLE);
		glLineStipple(1, 0xAAAA);
		glBegin(GL_LINES);
			for (float x = -WORLD_MESH; x <= WORLD_MESH; x ++) {
				glVertex2f(x, -WORLD_MESH);
				glVertex2f(x, +WORLD_MESH);
			}
			for (float y = -WORLD_MESH; y <= WORLD_MESH; y ++) {
				glVertex2f(-WORLD_MESH, y);
				glVertex2f(+WORLD_MESH, y);
			}
		glEnd();
		glDisable(GL_LINE_STIPPLE);

		glColor4f(.8f, .8f, .8f, 1.f);
		glBegin(GL_LINES);
			glVertex2f(-WORLD_MESH, 0);
			glVertex2f(+WORLD_MESH, 0);

			glVertex2f(0, -WORLD_MESH);
			glVertex2f(0, +WORLD_MESH);
		glEnd();
	}


    if ((mode == GL_RENDER) && (selectId == 0)) {
		glMatrixMode(GL_PROJECTION);
			glPushMatrix();
			glLoadIdentity();
	   		gluOrtho2D(0, worldWidth, 0, worldHeight); 
	    glMatrixMode(GL_MODELVIEW);
			glPushMatrix();
			glLoadIdentity();
			
			glTranslatef(worldWidth -30, 38, 0);
			glRotatef(angleScene, 0, 0, 1);

		glColor3f(1.f, 0.f, 0.f);
		glBegin(GL_LINES);
			glVertex2f(1.0, 0.0);
			glVertex2f(15.0, 0.0);
		glEnd();
		glRasterPos2f(20, 0);
		printfGlutBitmapFont(GLUT_BITMAP_HELVETICA_10, "%c", 'x');

		glColor3f(0.f, 1.f, 0.f);
		glBegin(GL_LINES);
			glVertex2f(1.0, 0.0);
			glVertex2f(0.0, 15.0);
		glEnd();
		glRasterPos2f(0, 20);
		printfGlutBitmapFont(GLUT_BITMAP_HELVETICA_10, "%c", 'y');

		glMatrixMode(GL_PROJECTION);
			glPopMatrix();
		glMatrixMode(GL_MODELVIEW);
			glPopMatrix();
	}

}


//-------------------------------------------------------------------------
//
//-------------------------------------------------------------------------
#define AREA_WIDTH  90
#define AREA_HEIGHT  12

static GLfloat wcx = 0, wcy = 0;
typedef enum { SAVE, LOAD } coordMode;

void worldCoord(coordMode mode) {
	static GLbyte buffer[4*sizeof(GLbyte) *AREA_WIDTH*AREA_HEIGHT];
    
	glMatrixMode(GL_PROJECTION);
		glPushMatrix();	glLoadIdentity();
   		gluOrtho2D(0, worldWidth, 0, worldHeight); 
    glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();
	
	if (mode == SAVE) 
		glReadPixels(worldWidth-AREA_WIDTH, 3,
			AREA_WIDTH, AREA_HEIGHT, GL_RGB, GL_BYTE, buffer);
	
	if (mode == LOAD) { 
		glDrawBuffer(GL_FRONT);
		memset(buffer, 0, 4*sizeof(GLbyte) *AREA_WIDTH*AREA_HEIGHT);////////////////
		glRasterPos2f(worldWidth-AREA_WIDTH, 3);
		glDrawPixels(AREA_WIDTH, AREA_HEIGHT, GL_RGB, GL_BYTE, buffer);
	}
	
	glColor4f(1.f, 1.f, .0f, 1.f);
	glRasterPos2f((float)(worldWidth-AREA_WIDTH), 5.f);
	printfGlutBitmapFont(GLUT_BITMAP_HELVETICA_10, 
		"x=%6.2f y=%6.2f", wcx, wcy);
	
	if (mode == LOAD)	glDrawBuffer(GL_BACK);

	glMatrixMode(GL_PROJECTION);
		glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
		glPopMatrix();
}


//-------------------------------------------------------------------------
//
//-------------------------------------------------------------------------
static void worldReshape(int width,  int height) 
{
	worldWidth = width; worldHeight = height;
	glViewport(0, 0, width, height); 
    glClearColor(0.f, 0.f, 0.f, 1.f);
	
	glClearStencil(0); 
//		glEnable(GL_STENCIL_TEST);
//		glStencilFunc(GL_EQUAL, 0, ~0);

	glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
   		//gluOrtho2D(-4, 4, -4, 4);
		glOrtho( -4.0, 4.0, -4.0, 4.0, -4.0,4.0);
}



static void worldDisplay(void) {
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); // vymazani barvoveho a hloubkoveho bufferu
    glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		glScalef(scaleScene, scaleScene, 1);
		glRotatef(angleScene, 0, 0, 1);
		glTranslatef(translScene[0], translScene[1], 0);
	
	worldDisplay(GL_RENDER); 
	glFlush();

	if (worldActive) worldCoord(SAVE);
	drawNurbs(GL_RENDER);
    glutSwapBuffers();
	
	unProject2D((GLfloat)worldMouseX, (GLfloat)(worldHeight-worldMouseY), &wcx, &wcy); //???????
}

//***************************************************************************	
//***************************************************************************	
//---------------------------------------------------------------------
// Tato funkce je volana pri posunu mysi bez stlaceneho tlacitka mysi
//---------------------------------------------------------------------

void worldMouseFunc(int button, int state, int x, int y) {
	unProject2D((GLfloat)x, (GLfloat)(worldHeight-y), &wcx, &wcy); 
	worldMouseX = x; worldMouseY = y;

	modifiers = glutGetModifiers();
	mouseStatus = button;

	switch (button) {
		case GLUT_LEFT_BUTTON:
			if (state == GLUT_DOWN) 
				switch (operations) {
					case ADD_POINT:	addCtrlPoint(wcx, wcy, 0, 1); 
						reDrawAll(); break;
					case DELETE_POINT: pickNurbs();
						if (worldHitId != -1) deleteCtrlPoint(worldHitId); 
						reDrawAll(); break;
					default: pickNurbs(); break; //PROC VOLA PICKNURBS, KDYZ NEZPRACUJE PICK???
				}
			break;
		case GLUT_MIDDLE_BUTTON: 
			break;
	}
}


void worldPassiveMotion(int x, int y) {
	static int hit = -1;

	unProject2D((GLfloat)x, (GLfloat)(worldHeight-y), &wcx, &wcy); 
	worldCoord(LOAD);
	worldMouseX = x; worldMouseY = y;
	if (pickNurbs(), hit != worldHitId) {
		static GLubyte buffer[] = { 0 };
		glDrawBuffer(GL_FRONT);
		glColor4f(1, (GLfloat)worldHitId /(numberOfKnots-order+1), 0, 1);
		//glRasterPos2f(ctrlPoints[worldHitId].x, ctrlPoints[worldHitId].y); // ORIGINAL
		//glRasterPos4f(ctrlPoints[i].x,ctrlPoints[i].y,ctrlPoints[i].z,ctrlPoints[i].w); // tohle nechodi, musim "RUCNE"	
		glRasterPos4f( 
			ctrlPoints[worldHitId].x/ctrlPoints[worldHitId].w,
			ctrlPoints[worldHitId].y/ctrlPoints[worldHitId].w,
			ctrlPoints[worldHitId].z/ctrlPoints[worldHitId].w,
			1.0);
		glBitmap(0, 0, 0, 0, -10, 6, buffer);
	
		if (worldHitId != -1) {
			printfGlutBitmapFont(GLUT_BITMAP_HELVETICA_10, 
				"w=%6.2f", ctrlPoints[worldHitId].w);
		} else
			glutPostRedisplay();

		glDrawBuffer(GL_BACK);
		hit = worldHitId;
	}
}


void worldMotionFunc(int x, int y) {
	switch (mouseStatus) {
		
		case GLUT_MIDDLE_BUTTON:    // v bodech meni vahy w
			if ((worldHitId != -1) || (modifiers == GLUT_ACTIVE_ALT)) {
				GLfloat xx, yy;
				unProject2D((GLfloat)x, (GLfloat)(worldHeight-y), &xx, &yy);
				printf("%.3f  %.3f\n", xx - wcx, yy - wcy);
				if (worldHitId != -1) {
					//GLfloat res = ctrlPoints[worldHitId].w 
					//	+ (sqr(ctrlPoints[worldHitId].w) +0.1) *(yy - wcy);
					//if ((-20 < res) && (res < 20)) 
					GLfloat w_new = ctrlPoints[worldHitId].w  + (GLfloat)(worldMouseY-y) * 0.1;
					if ((0 < w_new) && (w_new < 20)) 
					{
						float koef = w_new / ctrlPoints[worldHitId].w;
						ctrlPoints[worldHitId].w = w_new;
						//ctrlPoints[worldHitId].x *= koef;
						ctrlPoints[worldHitId].y *= koef;
						ctrlPoints[worldHitId].x *= koef;
					}
				}else {				// sem to nejde nikdy			
					translScene[0] += xx - wcx;
					translScene[1] += yy - wcy;
				}
				//wcx -= xx - wcx; wcy -= yy - wcy;
				worldMouseX = x; worldMouseY = y;
			}else {			// mimo body dela zoom sceny
				scaleScene += (y -worldMouseY) / (300.f); 
				worldReshape(worldWidth, worldHeight);
			}
			glutPostRedisplay();
			break;
								 
		case GLUT_LEFT_BUTTON: 
			if (worldHitId != -1) {
				GLfloat xx, yy;
				unProject2D((GLfloat)x, (GLfloat)(worldHeight-y), &xx, &yy);
				//moveCtrlPoint(worldHitId, wcx, wcy, 0, ctrlPoints[worldHitId].w);
				moveCtrlPoint(worldHitId, wcx, wcy, 0, ctrlPoints[worldHitId].w);
			} else 
				if (modifiers == GLUT_ACTIVE_ALT) 
					angleScene += 360.f/worldWidth  * (x -worldMouseX);
			glutPostRedisplay();
			break;

	}

	worldMouseX = x; worldMouseY = y;
}

void worldEntryFunc(int state) 
{
	if (state == GLUT_ENTERED) worldActive = true;
	else worldActive = false;
}


void worldKeyboardFunc(unsigned char k, int x, int y) { 
	int i = glutGetModifiers();
}


void applyWeights( vertex4f points[], int size )
{	
	int base = 0;
	for( int i=0; i < size; i++ )
	{
		float w = points[i].w;
		points[i].x *= w;
		points[i].y *= w;
		points[i].z *= w;
	}
}
#define SQ2 0.70710678118654752440084436210485   // sqrt(2)/2
void worldMenu(int polozka) 
{
	switch (polozka) {
		case 1:	operations = ADD_POINT; 
			addCtrlPoint(wcx, wcy, 0, 1); 
			reDrawAll();
			break;
		case 2:	operations = DELETE_POINT;
			if (pickNurbs(), worldHitId != -1) {
				deleteCtrlPoint(worldHitId); 
				//glutPostRedisplay(); 
			}
			break;
		case 3:	operations = EDIT_MODE; drawCtrlPolygon = !drawCtrlPolygon; break;
		case 4: operations = EDIT_MODE; break;

		case 10:  // new world
			numberOfCtrlPoints = 0;
			numberOfKnots = 2;
			knotVector[0] = 0; 
			knotVector[1] = 1;
			order = 2;

			reDrawAll();
			break;
		case 11: {  //DEMO 1 - 7-point circle
			vertex4f ctrlP[7 *sizeof(vertex4f)] =
				{	{ 1, 0, 0,   1}, { 1, 1, 0,.5},
					{-1, 1, 0,.5}, {-1, 0, 0,   1}, {-1,-1, 0,.5},
					{ 1,-1, 0,.5}, { 1, 0, 0,   1}	};
			applyWeights( ctrlP, 7 ); // new PF
			GLfloat knotV[12] =  { .0, .0, .0, .25, .5, .5,  .75, 1., 1., 1. };

			memcpy(ctrlPoints, ctrlP, 7 *sizeof(vertex4f));
			memcpy(knotVector, knotV, 10 *sizeof(GLfloat));
			numberOfCtrlPoints = 7;
			numberOfKnots = 10;
			order = 3;
			
			scaleScene = 1; angleScene = 0;
			translScene[0] = .0f;
			translScene[1] = .0f;
			reDrawAll();
			break;
		}
		case 12: {  //DEMO 2 - 9-point circle
			vertex4f ctrlP[9 *sizeof(vertex4f)] =
				{	{ 1, 0, 0,   1}, { 1, 1, 0,.707}, { 0, 1, 0,   1},
					{-1, 1, 0,.707}, {-1, 0, 0,   1}, {-1,-1, 0,.707},
					{ 0,-1, 0,   1}, { 1,-1, 0,.707}, { 1, 0, 0,   1}	};
			applyWeights( ctrlP, 9 ); // new PF
			GLfloat knotV[12] =  { .0, .0, .0, .25, .25, .5, .5, .75, .75, 1., 1., 1. };

			memcpy(ctrlPoints, ctrlP, 9 *sizeof(vertex4f));
			memcpy(knotVector, knotV, 12 *sizeof(GLfloat));
			numberOfCtrlPoints = 9;
			numberOfKnots = 12;
			order = 3;
			
			scaleScene = 1; angleScene = 0;
			translScene[0] = .0f;
			translScene[1] = .0f;
			reDrawAll();
			break;
		}
		case 13: {   // DEMO 2 - polygon
			vertex4f ctrlP[5 *sizeof(vertex4f)] =
				{	{-2, -1, 0, 1}, {-1, 1, 0, 1}, { 1, 1.5, 0, 1},
					{ 0, -1, 0, 1}, { 2, 0, 0, 1}	};
			applyWeights( ctrlP, 5 ); // new PF
			GLfloat knotV[8] =  { .0, .0, .0,   .33, .67,   1., 1., 1. };

			memcpy(ctrlPoints, ctrlP, 5 *sizeof(vertex4f));
			memcpy(knotVector, knotV, 8 *sizeof(GLfloat));
			numberOfCtrlPoints = 5;
			numberOfKnots = 8;
			order = 3;
			
			scaleScene = 1; angleScene = 0;
			translScene[0] = .0f;
			translScene[1] = .0f;
			reDrawAll();
			break;
		}
	}
    glutPostRedisplay(); //PF
}


void worldInit(void) {
	worldWindow = glutCreateSubWindow(mainWindow, 
	GAP, GAP, worldWidth, worldHeight);
	glutReshapeFunc(worldReshape);
	glutDisplayFunc( (void (*)(void))worldDisplay);
	glutPassiveMotionFunc(worldPassiveMotion);
	glutMouseFunc(worldMouseFunc);
	glutMotionFunc(worldMotionFunc);
	glutEntryFunc(worldEntryFunc);
	glutKeyboardFunc(worldKeyboardFunc);

	int worldMenuId = glutCreateMenu(worldMenu);
	glutAddMenuEntry("Add point   ",  1);
	glutAddMenuEntry("Delete point",  2);
	glutAddMenuEntry("Edit mode    ",  4);
	glutAddMenuEntry("Show Polygon",  3);
	glutAddMenuEntry("------------", -1);
	glutAddMenuEntry("New world   ", 10);
	glutAddMenuEntry("Demo 1 - 7 point circle", 11);
	glutAddMenuEntry("Demo 2 - 9 point circle", 12);
	glutAddMenuEntry("Demo 3 - polygon     ", 13);

	glutAttachMenu(GLUT_RIGHT_BUTTON);

	glutSetCursor(GLUT_CURSOR_FULL_CROSSHAIR);
}